#include <biu/RNAStructure.hh>
#include <stack>
#include <biu/assertbiu.hh>
#include <limits>


namespace biu
{
	const size_t RNAStructure::MIN_LOOP_LENGTH = 3;
	const size_t RNAStructure::INVALID_INDEX =
		std::numeric_limits<size_t>::max();
	const Alphabet RNAStructure::STRUCT_ALPH("().", 1);
	const Alphabet::AlphElem	RNAStructure::STRUCT_BND_OPEN = RNAStructure::STRUCT_ALPH.getElement("(");
	const Alphabet::AlphElem	RNAStructure::STRUCT_BND_CLOSE = RNAStructure::STRUCT_ALPH.getElement(")");
	const Alphabet::AlphElem	RNAStructure::STRUCT_UNBOUND = RNAStructure::STRUCT_ALPH.getElement(".");

//////////////////////////////////////////////////////////
// RNAStructure
//////////////////////////////////////////////////////////

	RNAStructure::RNAStructure(	const std::string& rnaSeqStr,
	 							const std::string& rnaStructBracketDotStr,
	 							const AllowedBasePairs* const _bPair)
	 :	rnaSeq(NULL), 
	 	rnaStructBracketDot(NULL),
	 	rnaLength(rnaSeqStr.size()),
		bPair(_bPair) 
	{
		assertbiu(bPair != NULL, "bPair is not allowed to be NULL.");
		assertbiu(bPair->getAlphabet()->isAlphabetString(rnaSeqStr),
		 "RNA sequence contains characters which are not in the alphabet.");
		assertbiu(STRUCT_ALPH.isAlphabetString(rnaStructBracketDotStr),
		 "RNA structure contains characters which are not in the alphabet.");

		// std::stack openingBonds contains opening bonds to create
		// vector rnaBonds during intitialisation
		std::stack<size_t>	openingBonds;
		
		rnaSeq = new Sequence(bPair->getAlphabet()->getSequence(rnaSeqStr));
		rnaStructBracketDot = new Structure(
		 STRUCT_ALPH.getSequence(rnaStructBracketDotStr));

		assertbiu(rnaSeq->size() == rnaStructBracketDot->size(),
		 "RNA sequence and structure have to have the same length.");
	
		rnaBonds = std::vector<size_t>
			   (rnaStructBracketDot->size(), RNAStructure::INVALID_INDEX);
		for (size_t i=0; i <  rnaStructBracketDot->size(); i++) {
			if ( (*rnaStructBracketDot)[i] == STRUCT_BND_OPEN ) {
				openingBonds.push(i);
			}
			else if ( (*rnaStructBracketDot)[i] == STRUCT_BND_CLOSE ) {
				assertbiu(!openingBonds.empty(),
				 "Closing without opening bond.");
				// check whether pairing of bases allowed	
				rnaBonds[openingBonds.top()] = i;
				openingBonds.pop();
			}
		}
		assertbiu(openingBonds.empty(), "Opening without closing bond.");
	}

	RNAStructure::RNAStructure(const RNAStructure& rnaStruct)
	 : rnaSeq(new Sequence(*rnaStruct.rnaSeq)),
	   rnaStructBracketDot(new Structure(*rnaStruct.rnaStructBracketDot)),
	   rnaLength(rnaStruct.rnaLength),
	   rnaBonds(rnaStruct.rnaBonds), bPair(rnaStruct.bPair) 
	{
		assertbiu(bPair != NULL, "bPair is not allowed to be NULL.");
		assertbiu(rnaSeq->size() == rnaStructBracketDot->size(),
		 "RNA sequence and structure have to have the same length.");
	}

	RNAStructure::~RNAStructure() {
		delete rnaSeq; rnaSeq = NULL;
		delete rnaStructBracketDot; rnaStructBracketDot = NULL;
	}

	bool 		
	RNAStructure::isAllowedBasePair(	size_t first,
			 							size_t second)  const 
	{
		if (first == RNAStructure::INVALID_INDEX || second == RNAStructure::INVALID_INDEX)
			return false;
		assertbiu(first < rnaLength && second < rnaLength,
		 "First or second is not a base position in the RNA sequence.");
		return bPair->allowedBasePair((*rnaSeq)[first], (*rnaSeq)[second]);
	}

	RNAStructure& 	
	RNAStructure::operator= (const RNAStructure& rnaStruct2) {
		if (this != &rnaStruct2) {
			if (rnaSeq->size() != rnaStruct2.rnaSeq->size()) {
				rnaSeq->resize(rnaStruct2.rnaSeq->size());
				rnaStructBracketDot->resize(rnaStruct2.rnaStructBracketDot->size());
			}
			std::copy(	rnaStruct2.rnaSeq->begin(),
						rnaStruct2.rnaSeq->end(),
						rnaSeq->begin());
			std::copy(	rnaStruct2.rnaStructBracketDot->begin(),
						rnaStruct2.rnaStructBracketDot->end(),
						rnaStructBracketDot->begin());
			bPair = rnaStruct2.bPair;
			rnaLength = rnaStruct2.rnaLength;	
			rnaBonds = rnaStruct2.rnaBonds;
		}
		return *this;
	}

	bool	 	
	RNAStructure::operator== (const RNAStructure& rnaStruct2)
			 const {
		return *rnaSeq == *(rnaStruct2.rnaSeq)
			&& *rnaStructBracketDot == *(rnaStruct2.rnaStructBracketDot)
			&& rnaBonds == rnaStruct2.rnaBonds
			&& *bPair == *(rnaStruct2.bPair);
	}

	bool	 	
	RNAStructure::operator!= (const RNAStructure& rnaStruct2)
			 const {
		return *rnaSeq != *(rnaStruct2.rnaSeq)
			|| *rnaStructBracketDot != *(rnaStruct2.rnaStructBracketDot)
			|| rnaBonds != rnaStruct2.rnaBonds
			|| *bPair != *(rnaStruct2.bPair);
	}

	bool		
	RNAStructure::hasValidStructure() const {
		int count = 0;	// counter for open bonds
		// each opening bracket has to have a closing bracket
		for (size_t i=0; i < rnaLength; i++) {
			if ((*rnaStructBracketDot)[i] == STRUCT_BND_OPEN ) {
				count++;
			}
			else if ((*rnaStructBracketDot)[i] == STRUCT_BND_CLOSE) {
				count--;
			}
			if (count < 0) {
				return false;
			}
		}
		return (count == 0);
	}

	bool		
	RNAStructure::hasValidBasePairs() const{
		for (size_t i=0; i < rnaBonds.size(); i++) {
			// existing bond has to be an allowed base pair
			if (rnaBonds[i] != RNAStructure::INVALID_INDEX
			    && !isAllowedBasePair(i, rnaBonds[i])) {
				return false;
			}
		}
		return true;
	}	

	bool		
	RNAStructure::hasMinLoopLength() const {
		for (size_t i=0; i < rnaBonds.size(); i++) {
			// existing bond has to have at least MIN_LENGTH_LOOP
			if (rnaBonds[i] != RNAStructure::INVALID_INDEX
			    && (rnaBonds[i]-i) <= RNAStructure::MIN_LOOP_LENGTH) {
				return false;
			}
		}
		return true;
	}


	bool		
	RNAStructure::isValid() const{
		assertbiu(rnaSeq->size() == rnaStructBracketDot->size(),
                 "RNA sequence and structure have to have the same length.");
		return (	hasValidStructure() && hasMinLoopLength()
					&& hasValidBasePairs() );
	}

	std::string	
	RNAStructure::getStringRepresentation() const {
		return ( getStructureString() + "(" + getSequenceString() + ")" );
	}

} // namespace biu
